# -*- coding: utf-8 -*-
"""Implementation of :class:`aiida.orm.implementation.storage_backend.StorageBackend` using PostgreSQL + Azure."""
from __future__ import annotations

import typing as t

from ..repository.azure_blob import AzureBlobStorageRepositoryBackend
from .psql_dos import BasePsqlDosBackend
from .psql_s3 import PsqlS3StorageMigrator


class PsqlAzureBlobStorageMigrator(PsqlS3StorageMigrator):
    """Subclass :class:`aiida.storage.psql_dos.migrator.PsqlDosMigrator` to customize the repository implementation."""

    def reset_repository(self) -> None:
        """Reset the repository deleting the bucket and all its contents."""
        repository = self.get_repository()
        if repository.is_initialised:
            repository.delete_objects(repository.list_objects())

    def get_repository(self) -> AzureBlobStorageRepositoryBackend:  # type: ignore[override]
        """Return the file repository backend instance.

        :returns: The repository of the configured profile.
        """
        storage_config = self.profile.storage_config

        return AzureBlobStorageRepositoryBackend(
            container_name=storage_config['container_name'],
            connection_string=storage_config['connection_string'],
        )


class PsqlAzureBlobStorage(BasePsqlDosBackend):
    """Storage backend using PostgresSQL and Azure Blob Storage."""

    migrator = PsqlAzureBlobStorageMigrator

    def get_repository(self) -> AzureBlobStorageRepositoryBackend:  # type: ignore[override]
        """Return the file repository backend instance.

        :returns: The repository of the configured profile.
        """
        return self.migrator(self.profile).get_repository()

    @classmethod
    def _get_cli_options(cls) -> dict[str, t.Any]:
        """Return the CLI options that would allow to create an instance of this class."""
        options = super()._get_cli_options()
        options.update(
            **{
                'container_name': {
                    'required': True,
                    'type': str,
                    'prompt': 'Container name',
                    'help': 'The Azure Blob Storage container name.',
                },
                'connection_string': {
                    'required': True,
                    'type': str,
                    'prompt': 'Connection string',
                    'help': 'The Azure Blob Storage connection string.',
                }
            }
        )
        return options



# -*- coding: utf-8 -*-
# pylint: disable=cyclic-import,wrong-import-position
"""Command line interface for ``aiida-s3``."""
from aiida.cmdline.groups.verdi import VerdiCommandGroup
import click


@click.group('aiida-s3', cls=VerdiCommandGroup)
def cmd_root():
    """Command line interface for ``aiida-s3``."""


from .cmd_profile import cmd_profile



# -*- coding: utf-8 -*-
"""Implementation of the :py:`aiida.repository.backend.abstract.AbstractRepositoryBackend` using Azure Blob Storage."""
from __future__ import annotations

import contextlib
import tempfile
import typing as t
import uuid

from aiida.repository.backend.abstract import AbstractRepositoryBackend
from azure.storage.blob import BlobServiceClient, ContainerClient

__all__ = ('AzureBlobStorageRepositoryBackend',)


class AzureBlobStorageRepositoryBackend(AbstractRepositoryBackend):
    """Implementation of the ``AbstractRepositoryBackend`` using Azure Blob Storage as the backend."""

    def __init__(self, container_name: str, connection_string: str):
        """Construct a new instance for a given storage account and container.

        .. note:: It is possible to construct an instance for a container that doesn't exist yet. To use the backend,
            however, the container needs to exist. To ensure it exists, call ``initialise``, which will create the
            container if it doesn't already.

        :param container_name: The name of the container to use.
        :param connection_string: The connection string for the Azure Blob Storage storage account.
        """
        self._container_name: str = container_name
        self._connection_string: str = connection_string

        try:
            self._service_client: BlobServiceClient = BlobServiceClient.from_connection_string(self._connection_string)
        except ValueError as exception:
            raise ValueError(
                f'could not connect with the given connection string: {self._connection_string}'
            ) from exception

        if self._service_client is None:
            raise ValueError(f'failed to connect with the given connection string: {self._connection_string}')

        self._container_client: ContainerClient = self._service_client.get_container_client(self._container_name)

    def __str__(self) -> str:
        """Return the string representation of this repository."""
        return f'AwsS3RepositoryBackend: <{self._container_name}>'

    @property
    def _container_exists(self) -> bool:
        """Return whether the container exists."""
        return self._container_client.exists()

    @property
    def is_initialised(self) -> bool:
        """Return whether the repository has been initialised.

        This amounts to whether the configured container actually exists. Calling ``initialise`` will create the
        container if it didn't already exist.
        """
        return self._container_exists

    def initialise(self, **kwargs) -> None:
        """Initialise the repository if it hasn't already been initialised.

        :param kwargs: parameters for the initialisation.
        """
        if self.is_initialised:
            return

        self._container_client.create_container(**kwargs)

    @property
    def uuid(self) -> str:
        """Return the unique identifier of the repository."""
        return self._container_name

    @property
    def key_format(self) -> str:
        """Return the format for the keys of the repository."""
        return 'uuid4'

    def erase(self):
        """Delete the container configured for this instance and all its contents."""
        if not self._container_exists:
            return

        self._container_client.delete_container()

    def _put_object_from_filelike(self, handle: t.BinaryIO) -> str:
        """Store the byte contents of a file in the repository.

        :param handle: filelike object with the byte content to be stored.
        :return: the generated fully qualified identifier for the object within the repository.
        :raises TypeError: if the handle is not a byte stream.
        """
        key = str(uuid.uuid4())
        self._container_client.upload_blob(name=key, data=handle)
        return key

    def has_objects(self, keys: list[str]) -> list[bool]:
        """Return whether the repository has an object with the given key.

        :param keys: list of fully qualified identifiers for objects within the repository.
        :return: list of booleans, in the same order as the keys provided, with value True if the respective object
            exists and False otherwise.
        """
        existing_keys = set(self.list_objects())
        return [key in existing_keys for key in keys]

    @contextlib.contextmanager
    def open(self, key: str) -> t.Iterator[t.IO[bytes]]:  # type: ignore[override]
        """Open a file handle to an object stored under the given key.

        .. note:: this should only be used to open a handle to read an existing file. To write a new file use the method
            ``put_object_from_filelike`` instead.

        :param key: fully qualified identifier for the object within the repository.
        :return: yield a byte stream object.
        :raise FileNotFoundError: if the file does not exist.
        :raise OSError: if the file could not be opened.
        """
        with tempfile.TemporaryFile(mode='w+b') as handle:
            try:
                self._container_client.download_blob(key).readinto(handle)
            except Exception as exception:  # pylint: disable=broad-except
                raise FileNotFoundError(f'object with key `{key}` does not exist.') from exception
            handle.seek(0)
            yield handle

    def iter_object_streams(self, keys: list[str]) -> t.Iterator[tuple[str, t.IO[bytes]]]:  # type: ignore[override]
        """Return an iterator over the (read-only) byte streams of objects identified by key.

        .. note:: handles should only be read within the context of this iterator.

        :param keys: fully qualified identifiers for the objects within the repository.
        :return: an iterator over the object byte streams.
        :raise FileNotFoundError: if the file does not exist.
        :raise OSError: if a file could not be opened.
        """
        for key in keys:
            with self.open(key) as handle:
                yield key, handle

    def delete_objects(self, keys: t.Iterable[str]) -> None:
        """Delete the objects from the repository.

        :param keys: list of fully qualified identifiers for the objects within the repository.
        :raise FileNotFoundError: if any of the files does not exist.
        :raise OSError: if any of the files could not be deleted.
        """
        super().delete_objects(list(keys))
        if keys:
            self._container_client.delete_blobs(*keys)

    def list_objects(self) -> t.Iterable[str]:
        """Return iterable that yields all available objects by key.

        :return: An iterable for all the available object keys.
        """
        for name in self._container_client.list_blob_names():
            yield name

    def maintain(  # type: ignore[override]
        self,
        dry_run: bool = False,
        live: bool = True,
        pack_loose: bool = None,
        do_repack: bool = None,
        clean_storage: bool = None,
        do_vacuum: bool = None,
    ) -> dict:
        # pylint: disable=arguments-differ, unused-argument
        """Perform maintenance operations.

        :param live: if True, will only perform operations that are safe to do while the repository is in use.
        :param pack_loose: flag for forcing the packing of loose files.
        :param do_repack: flag for forcing the re-packing of already packed files.
        :param clean_storage: flag for forcing the cleaning of soft-deleted files from the repository.
        :param do_vacuum: flag for forcing the vacuuming of the internal database when cleaning the repository.
        :return: a dictionary with information on the operations performed.
        """
        return {}

    def get_info(  # type: ignore[override]
        self,
        detailed=False,
    ) -> dict[str, int | str | dict[str, int] | dict[str, float]]:
        # pylint: disable=arguments-differ, unused-argument
        """Return information on configuration and content of the repository."""
        return {}



# -*- coding: utf-8 -*-
"""Implementation of the :py:`aiida.repository.backend.abstract.AbstractRepositoryBackend` using AWS S3 as backend."""
from __future__ import annotations

import boto3

from .s3 import S3RepositoryBackend

__all__ = ('AwsS3RepositoryBackend',)


class AwsS3RepositoryBackend(S3RepositoryBackend):
    """Implementation of the ``AbstractRepositoryBackend`` using S3 as the backend."""

    def __init__(self, aws_access_key_id: str, aws_secret_access_key: str, region_name: str, bucket_name: str):
        """Construct a new instance for a given bucket.

        .. note:: It is possible to construct an instance for a bucket that doesn't exist yet. To use the backend,
            however, the bucket needs to exist. To ensure it exists, call ``initialise``, which will create the bucket
            if it doesn't already.

        :param aws_access_key_id: The AWS access key ID to use to authenticate with AWS S3.
        :param aws_secret_access_key: The AWS secret access key to use to authenticate with AWS S3.
        :param region_name: The AWS region name to create the bucket in if it doesn't yet exist.
        :param bucket_name: The name of the bucket to use.
        """
        # pylint: disable=super-init-not-called
        self._bucket_name = bucket_name
        self._region_name = region_name
        self._client = boto3.client(
            's3',
            aws_access_key_id=aws_access_key_id,
            aws_secret_access_key=aws_secret_access_key,
            region_name=region_name,
        )

    def __str__(self) -> str:
        """Return the string representation of this repository."""
        return f'AwsS3RepositoryBackend: <{self._bucket_name}>'

    def initialise(self, **kwargs) -> None:
        """Initialise the repository if it hasn't already been initialised.

        :param kwargs: parameters for the initialisation.
        """
        if self._region_name is not None:
            kwargs['CreateBucketConfiguration'] = {'LocationConstraint': self._region_name}

        super().initialise(**kwargs)



# -*- coding: utf-8 -*-
"""Implementation of :class:`aiida.orm.implementation.storage_backend.StorageBackend` using PostgreSQL + AWS S3."""
from __future__ import annotations

import typing as t

from ..repository.aws_s3 import AwsS3RepositoryBackend
from .psql_dos import BasePsqlDosBackend
from .psql_s3 import PsqlS3StorageMigrator


class PsqlAwsS3StorageMigrator(PsqlS3StorageMigrator):
    """Subclass :class:`aiida.storage.psql_dos.migrator.PsqlDosMigrator` to customize the repository implementation."""

    def get_repository(self) -> AwsS3RepositoryBackend:
        """Return the file repository backend instance.

        :returns: The repository of the configured profile.
        """
        storage_config = self.profile.storage_config

        return AwsS3RepositoryBackend(
            aws_access_key_id=storage_config['aws_access_key_id'],
            aws_secret_access_key=storage_config['aws_secret_access_key'],
            region_name=storage_config['aws_region_name'],
            bucket_name=storage_config['aws_bucket_name'],
        )


class PsqlAwsS3Storage(BasePsqlDosBackend):
    """Storage backend using PostgresSQL and AWS S3."""

    migrator = PsqlAwsS3StorageMigrator

    def get_repository(self) -> AwsS3RepositoryBackend:  # type: ignore[override]
        """Return the file repository backend instance.

        :returns: The repository of the configured profile.
        """
        return self.migrator(self.profile).get_repository()

    @classmethod
    def _get_cli_options(cls) -> dict[str, t.Any]:
        """Return the CLI options that would allow to create an instance of this class."""
        options = super()._get_cli_options()
        options.update(
            **{
                'aws_access_key_id': {
                    'required': True,
                    'type': str,
                    'prompt': 'AWS access key ID',
                    'help': 'The AWS access key ID.',
                },
                'aws_secret_access_key': {
                    'required': True,
                    'type': str,
                    'prompt': 'AWS secret access key',
                    'help': 'The AWS secret access key.',
                },
                'aws_region_name': {
                    'required': True,
                    'type': str,
                    'prompt': 'AWS region',
                    'help': 'The AWS region name code, e.g., `eu-central-1`.',
                },
                'aws_bucket_name': {
                    'required': True,
                    'type': str,
                    'prompt': 'AWS bucket name',
                    'help': 'The name of the AWS S3 bucket to use.',
                },
            }
        )
        return options



# -*- coding: utf-8 -*-
"""Implementation of :class:`aiida.orm.implementation.storage_backend.StorageBackend` using PostgreSQL + AWS S3."""
from __future__ import annotations

import collections

from aiida.plugins.entry_point import EntryPoint
from aiida.storage.psql_dos import PsqlDosBackend


class BasePsqlDosBackend(PsqlDosBackend):
    """Storage backend using PostgresSQL and AWS S3."""

    @classmethod
    def get_entry_point(cls) -> EntryPoint | None:
        """Return the entry point with which this storage backend implementation is registered.

        :return: The entry point or ``None`` if not found.
        """
        from aiida.plugins.entry_point import get_entry_point_from_class
        return get_entry_point_from_class(cls.__module__, cls.__name__)[1]

    @classmethod
    def get_cli_options(cls) -> collections.OrderedDict:
        """Return the CLI options that would allow to create an instance of this class."""
        return collections.OrderedDict(cls._get_cli_options())

    @classmethod
    def _get_cli_options(cls) -> dict:
        """Return the CLI options that would allow to create an instance of this class."""
        return {
            'profile_name': {
                'required': True,
                'type': str,
                'prompt': 'Profile name',
                'help': 'The name of the profile.',
            },
            'postgresql_engine': {
                'required': True,
                'type': str,
                'prompt': 'Postgresql engine',
                'default': 'postgresql_psycopg2',
                'help': 'The engine to use to connect to the database.',
            },
            'postgresql_hostname': {
                'required': True,
                'type': str,
                'prompt': 'Postgresql hostname',
                'default': 'localhost',
                'help': 'The hostname of the PostgreSQL server.',
            },
            'postgresql_port': {
                'required': True,
                'type': int,
                'prompt': 'Postgresql port',
                'default': '5432',
                'help': 'The port of the PostgreSQL server.',
            },
            'postgresql_username': {
                'required': True,
                'type': str,
                'prompt': 'Postgresql username',
                'help': 'The username with which to connect to the PostgreSQL server.',
            },
            'postgresql_password': {
                'required': True,
                'type': str,
                'prompt': 'Postgresql password',
                'help': 'The password with which to connect to the PostgreSQL server.',
            },
            'postgresql_database_name': {
                'required': True,
                'type': str,
                'prompt': 'Postgresql database name',
                'help': 'The name of the database in the PostgreSQL server.',
            }
        }



# -*- coding: utf-8 -*-
"""CLI commands to create profiles."""
from aiida.cmdline.groups import DynamicEntryPointCommandGroup

from . import cmd_root


@cmd_root.group('profile')  # type: ignore[has-type]
def cmd_profile():
    """Commands to create profiles."""


def create_profile(cls, non_interactive, **kwargs):  # pylint: disable=unused-argument
    """Set up a new profile with an ``aiida-s3`` storage backend."""
    import contextlib
    import io

    from aiida.cmdline.utils import echo
    from aiida.manage.configuration import Profile, get_config

    profile_name = kwargs.pop('profile_name')
    profile_config = {
        'storage': {
            'backend': cls.get_entry_point().name,
            'config': {
                'database_engine': kwargs.pop('postgresql_engine'),
                'database_hostname': kwargs.pop('postgresql_hostname'),
                'database_port': kwargs.pop('postgresql_port'),
                'database_username': kwargs.pop('postgresql_username'),
                'database_password': kwargs.pop('postgresql_password'),
                'database_name': kwargs.pop('postgresql_database_name'),
            }
        },
        'process_control': {
            'backend': 'rabbitmq',
            'config': {
                'broker_protocol': 'amqp',
                'broker_username': 'guest',
                'broker_password': 'guest',
                'broker_host': '127.0.0.1',
                'broker_port': 5672,
                'broker_virtual_host': ''
            }
        },
    }

    profile_config['storage']['config'].update(**kwargs)
    profile = Profile(profile_name, profile_config)

    config = get_config()

    if profile.name in config.profile_names:
        echo.echo_critical(f'The profile `{profile.name}` already exists.')

    config.add_profile(profile)

    echo.echo_report('Initialising the storage backend.')
    try:
        with contextlib.redirect_stdout(io.StringIO()):
            profile.storage_cls.initialise(profile)
    except Exception as exception:  # pylint: disable=broad-except
        echo.echo_critical(
            f'Storage backend initialisation failed, probably because connection details are incorrect:\n{exception}'
        )
    else:
        echo.echo_success('Storage backend initialisation completed.')

    config.store()
    echo.echo_success(f'Created new profile `{profile.name}`.')


@cmd_profile.group(
    'setup',
    cls=DynamicEntryPointCommandGroup,
    command=create_profile,
    entry_point_group='aiida.storage',
    entry_point_name_filter=r's3\..*'
)
def cmd_profile_setup():
    """Set up a new profile with an ``aiida-s3`` storage backend."""



# -*- coding: utf-8 -*-
"""Implementation of :class:`aiida.orm.implementation.storage_backend.StorageBackend` using PostgreSQL + S3."""
from __future__ import annotations

import typing as t

from aiida.storage.psql_dos.migrator import PsqlDosMigrator

from ..repository.s3 import S3RepositoryBackend
from .psql_dos import BasePsqlDosBackend


class PsqlS3StorageMigrator(PsqlDosMigrator):
    """Subclass :class:`aiida.storage.psql_dos.migrator.PsqlDosMigrator` to customize the repository implementation."""

    def get_repository_uuid(self) -> str:
        """Return the UUID of the repository.

        :returns: The UUID of the repository of the configured profile.
        """
        return self.get_repository().uuid

    def reset_repository(self) -> None:
        """Reset the repository deleting the bucket and all its contents."""
        repository = self.get_repository()
        if repository.is_initialised:
            repository.erase()

    def initialise_repository(self) -> None:
        """Initialise the repository."""
        repository = self.get_repository()
        repository.initialise()

    @property
    def is_repository_initialised(self) -> bool:
        """Return whether the repository is initialised.

        :returns: Boolean, ``True`` if the repository is initalised, ``False`` otherwise.
        """
        return self.get_repository().is_initialised

    def get_repository(self) -> S3RepositoryBackend:
        """Return the file repository backend instance.

        :returns: The repository of the configured profile.
        """
        storage_config = self.profile.storage_config

        return S3RepositoryBackend(
            endpoint_url=storage_config['endpoint_url'],
            access_key_id=storage_config['access_key_id'],
            secret_access_key=storage_config['secret_access_key'],
            bucket_name=storage_config['bucket_name'],
        )


class PsqlS3Storage(BasePsqlDosBackend):
    """Storage backend using PostgresSQL and S3 object store."""

    migrator = PsqlS3StorageMigrator

    def get_repository(self) -> S3RepositoryBackend:  # type: ignore[override]
        """Return the file repository backend instance.

        :returns: The repository of the configured profile.
        """
        return self.migrator(self.profile).get_repository()

    @classmethod
    def _get_cli_options(cls) -> dict[str, t.Any]:
        """Return the CLI options that would allow to create an instance of this class."""
        options = super()._get_cli_options()
        options.update(
            **{
                'endpoint_url': {
                    'required': True,
                    'type': str,
                    'prompt': 'Server endpoint URL',
                    'help': 'The endpoint URL of the server to connect to.',
                },
                'access_key_id': {
                    'required': True,
                    'type': str,
                    'prompt': 'Access key ID',
                    'help': 'The access key ID.',
                },
                'secret_access_key': {
                    'required': True,
                    'type': str,
                    'prompt': 'Secret access key',
                    'help': 'The secret access key.',
                },
                'bucket_name': {
                    'required': True,
                    'type': str,
                    'prompt': 'Bucket name',
                    'help': 'The name of the S3 bucket to use.',
                },
            }
        )
        return options



# -*- coding: utf-8 -*-
"""Implementation of the :py:`aiida.repository.backend.abstract.AbstractRepositoryBackend` using S3 as backend."""
from __future__ import annotations

import contextlib
import tempfile
import typing as t
import uuid

from aiida.repository.backend.abstract import AbstractRepositoryBackend
import boto3
import botocore

__all__ = ('S3RepositoryBackend',)


class S3RepositoryBackend(AbstractRepositoryBackend):
    """Implementation of the ``AbstractRepositoryBackend`` using S3 as the backend."""

    def __init__(self, endpoint_url: str, access_key_id: str, secret_access_key: str, bucket_name: str):
        """Construct a new instance for a given bucket.

        .. note:: It is possible to construct an instance for a bucket that doesn't exist yet. To use the backend,
            however, the bucket needs to exist. To ensure it exists, call ``initialise``, which will create the bucket
            if it doesn't already.

        :param endpoint_url: The endpoint URL of the server to connect to.
        :param access_key_id: The access key ID to use to authenticate with S3.
        :param secret_access_key: The secret access key to use to authenticate with S3.
        :param bucket_name: The name of the bucket to use.
        """
        self._bucket_name = bucket_name
        self._client = boto3.client(
            's3',
            endpoint_url=endpoint_url,
            aws_access_key_id=access_key_id,
            aws_secret_access_key=secret_access_key,
        )

    def __str__(self) -> str:
        """Return the string representation of this repository."""
        return f'S3RepositoryBackend: <{self._bucket_name}>'

    @property
    def _bucket_exists(self) -> bool:
        """Return whether the bucket exists."""
        try:
            self._client.head_bucket(Bucket=self._bucket_name)
        except botocore.exceptions.ClientError:
            return False

        return True

    @property
    def is_initialised(self) -> bool:
        """Return whether the repository has been initialised.

        This amounts to whether the configured bucket actually exists. Calling ``initialise`` will create the bucket
        if it didn't already exist.
        """
        return self._bucket_exists

    def initialise(self, **kwargs) -> None:
        """Initialise the repository if it hasn't already been initialised.

        :param kwargs: parameters for the initialisation.
        """
        if self.is_initialised:
            return

        self._client.create_bucket(Bucket=self._bucket_name, **kwargs)

    @property
    def uuid(self) -> str:
        """Return the unique identifier of the repository."""
        return self._bucket_name

    @property
    def key_format(self) -> str | None:
        """Return the format for the keys of the repository."""
        return 'uuid4'

    def erase(self):
        """Delete the bucket configured for this instance and all its contents."""
        if not self._bucket_exists:
            return

        self.delete_objects(list(self.list_objects()))
        self._client.delete_bucket(Bucket=self._bucket_name)

    def _put_object_from_filelike(self, handle: t.BinaryIO) -> str:
        """Store the byte contents of a file in the repository.

        :param handle: filelike object with the byte content to be stored.
        :return: the generated fully qualified identifier for the object within the repository.
        :raises TypeError: if the handle is not a byte stream.
        """
        key = str(uuid.uuid4())
        self._client.put_object(Bucket=self._bucket_name, Body=handle, Key=key)
        return key

    def has_objects(self, keys: list[str]) -> list[bool]:
        """Return whether the repository has an object with the given key.

        :param keys: list of fully qualified identifiers for objects within the repository.
        :return: list of booleans, in the same order as the keys provided, with value True if the respective object
            exists and False otherwise.
        """
        existing_keys = set(self.list_objects())
        return [key in existing_keys for key in keys]

    @contextlib.contextmanager
    def open(self, key: str) -> t.Iterator[t.IO[bytes]]:  # type: ignore[override]
        """Open a file handle to an object stored under the given key.

        .. note:: this should only be used to open a handle to read an existing file. To write a new file use the method
            ``put_object_from_filelike`` instead.

        :param key: fully qualified identifier for the object within the repository.
        :return: yield a byte stream object.
        :raise FileNotFoundError: if the file does not exist.
        :raise OSError: if the file could not be opened.
        """
        try:
            with tempfile.TemporaryFile() as handle:
                self._client.download_fileobj(self._bucket_name, key, handle)
                handle.seek(0)
                yield handle
        except botocore.exceptions.ClientError as exception:
            raise FileNotFoundError(f'object with key `{key}` does not exist.') from exception

    def iter_object_streams(self, keys: list[str]) -> t.Iterator[tuple[str, t.IO[bytes]]]:  # type: ignore[override]
        """Return an iterator over the (read-only) byte streams of objects identified by key.

        .. note:: handles should only be read within the context of this iterator.

        :param keys: fully qualified identifiers for the objects within the repository.
        :return: an iterator over the object byte streams.
        :raise FileNotFoundError: if the file does not exist.
        :raise OSError: if a file could not be opened.
        """
        for key in keys:
            with self.open(key) as handle:
                yield key, handle

    def delete_objects(self, keys: list[str]) -> None:
        """Delete the objects from the repository.

        :param keys: list of fully qualified identifiers for the objects within the repository.
        :raise FileNotFoundError: if any of the files does not exist.
        :raise OSError: if any of the files could not be deleted.
        """
        super().delete_objects(keys)
        if keys:
            self._client.delete_objects(Bucket=self._bucket_name, Delete={'Objects': [{'Key': key} for key in keys]})

    def list_objects(self) -> t.Iterable[str]:
        """Return iterable that yields all available objects by key.

        :return: An iterable for all the available object keys.
        """
        kwargs = {'Bucket': self._bucket_name}
        paginator = self._client.get_paginator('list_objects_v2')

        for page in paginator.paginate(**kwargs):
            try:
                contents = page['Contents']
            except KeyError:
                break

            for obj in contents:
                yield obj['Key']

    def maintain(  # type: ignore[override]
        self,
        dry_run: bool = False,
        live: bool = True,
        pack_loose: bool = None,
        do_repack: bool = None,
        clean_storage: bool = None,
        do_vacuum: bool = None,
    ) -> dict:
        # pylint: disable=arguments-differ, unused-argument
        """Perform maintenance operations.

        :param live: if True, will only perform operations that are safe to do while the repository is in use.
        :param pack_loose: flag for forcing the packing of loose files.
        :param do_repack: flag for forcing the re-packing of already packed files.
        :param clean_storage: flag for forcing the cleaning of soft-deleted files from the repository.
        :param do_vacuum: flag for forcing the vacuuming of the internal database when cleaning the repository.
        :return: a dictionary with information on the operations performed.
        """
        return {}

    def get_info(  # type: ignore[override]
        self,
        detailed=False,
    ) -> dict[str, int | str | dict[str, int] | dict[str, float]]:
        # pylint: disable=arguments-differ, unused-argument
        """Return information on configuration and content of the repository."""
        return {}
